#
# Copyright 2013-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Easyconfig templates module that provides templating that can
be used within an Easyconfig file.

:author: Stijn De Weirdt (Ghent University)
:author: Fotis Georgatos (Uni.Lu, NTUA)
:author: Kenneth Hoste (Ghent University)
"""
import re
import platform

from easybuild.base import fancylogger
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.py2vs3 import string_type
from easybuild.tools.systemtools import get_shared_lib_ext, pick_dep_version
from easybuild.tools.config import build_option


_log = fancylogger.getLogger('easyconfig.templates', fname=False)

# derived from easyconfig, but not from ._config directly
TEMPLATE_NAMES_EASYCONFIG = [
    ('module_name', "Module name"),
    ('nameletter', "First letter of software name"),
    ('toolchain_name', "Toolchain name"),
    ('toolchain_version', "Toolchain version"),
    ('version_major_minor', "Major.Minor version"),
    ('version_major', "Major version"),
    ('version_minor', "Minor version"),
]
# derived from EasyConfig._config
TEMPLATE_NAMES_CONFIG = [
    'bitbucket_account',
    'github_account',
    'name',
    'parallel',
    'version',
    'versionsuffix',
    'versionprefix',
]
# lowercase versions of ._config
TEMPLATE_NAMES_LOWER_TEMPLATE = "%(name)slower"
TEMPLATE_NAMES_LOWER = [
    'name',
    'nameletter',
]
# values taken from the EasyBlock before each step
TEMPLATE_NAMES_EASYBLOCK_RUN_STEP = [
    ('builddir', "Build directory"),
    ('installdir', "Installation directory"),
]
# software names for which to define <pref>ver and <pref>shortver templates
TEMPLATE_SOFTWARE_VERSIONS = [
    # software name, prefix for *ver and *shortver
    ('CUDA', 'cuda'),
    ('CUDAcore', 'cuda'),
    ('Java', 'java'),
    ('Perl', 'perl'),
    ('Python', 'py'),
    ('R', 'r'),
]
# template values which are only generated dynamically
TEMPLATE_NAMES_DYNAMIC = [
    ('arch', "System architecture (e.g. x86_64, aarch64, ppc64le, ...)"),
    ('mpi_cmd_prefix', "Prefix command for running MPI programs (with default number of ranks)"),
    ('cuda_compute_capabilities', "Comma-separated list of CUDA compute capabilities, as specified via "
     "--cuda-compute-capabilities configuration option or via cuda_compute_capabilities easyconfig parameter"),
    ('cuda_cc_space_sep', "Space-separated list of CUDA compute capabilities"),
    ('cuda_cc_semicolon_sep', "Semicolon-separated list of CUDA compute capabilities"),
    ('cuda_sm_comma_sep', "Comma-separated list of sm_* values that correspond with CUDA compute capabilities"),
    ('cuda_sm_space_sep', "Space-separated list of sm_* values that correspond with CUDA compute capabilities"),
]

# constant templates that can be used in easyconfigs
TEMPLATE_CONSTANTS = [
    # source url constants
    ('APACHE_SOURCE', 'https://archive.apache.org/dist/%(namelower)s',
     'apache.org source url'),
    ('BITBUCKET_SOURCE', 'https://bitbucket.org/%(bitbucket_account)s/%(namelower)s/get',
     'bitbucket.org source url (namelower is used if bitbucket_account easyconfig parameter is not specified)'),
    ('BITBUCKET_DOWNLOADS', 'https://bitbucket.org/%(bitbucket_account)s/%(namelower)s/downloads',
     'bitbucket.org downloads url (namelower is used if bitbucket_account easyconfig parameter is not specified)'),
    ('CRAN_SOURCE', 'https://cran.r-project.org/src/contrib',
     'CRAN (contrib) source url'),
    ('FTPGNOME_SOURCE', 'https://ftp.gnome.org/pub/GNOME/sources/%(namelower)s/%(version_major_minor)s',
     'http download for gnome ftp server'),
    ('GITHUB_SOURCE', 'https://github.com/%(github_account)s/%(name)s/archive',
     'GitHub source URL (namelower is used if github_account easyconfig parameter is not specified)'),
    ('GITHUB_LOWER_SOURCE', 'https://github.com/%(github_account)s/%(namelower)s/archive',
     'GitHub source URL (lowercase name, namelower is used if github_account easyconfig parameter is not specified)'),
    ('GNU_SAVANNAH_SOURCE', 'https://download-mirror.savannah.gnu.org/releases/%(namelower)s',
     'download.savannah.gnu.org source url'),
    ('GNU_SOURCE', 'https://ftpmirror.gnu.org/gnu/%(namelower)s',
     'gnu.org source url'),
    ('GOOGLECODE_SOURCE', 'http://%(namelower)s.googlecode.com/files',
     'googlecode.com source url'),
    ('LAUNCHPAD_SOURCE', 'https://launchpad.net/%(namelower)s/%(version_major_minor)s.x/%(version)s/+download/',
     'launchpad.net source url'),
    ('PYPI_SOURCE', 'https://pypi.python.org/packages/source/%(nameletter)s/%(name)s',
     'pypi source url'),  # e.g., Cython, Sphinx
    ('PYPI_LOWER_SOURCE', 'https://pypi.python.org/packages/source/%(nameletterlower)s/%(namelower)s',
     'pypi source url (lowercase name)'),  # e.g., Greenlet, PyZMQ
    ('R_SOURCE', 'https://cran.r-project.org/src/base/R-%(version_major)s',
     'cran.r-project.org (base) source url'),
    ('SOURCEFORGE_SOURCE', 'https://download.sourceforge.net/%(namelower)s',
     'sourceforge.net source url'),
    ('XORG_DATA_SOURCE', 'https://xorg.freedesktop.org/archive/individual/data/',
     'xorg data source url'),
    ('XORG_LIB_SOURCE', 'https://xorg.freedesktop.org/archive/individual/lib/',
     'xorg lib source url'),
    ('XORG_PROTO_SOURCE', 'https://xorg.freedesktop.org/archive/individual/proto/',
     'xorg proto source url'),
    ('XORG_UTIL_SOURCE', 'https://xorg.freedesktop.org/archive/individual/util/',
     'xorg util source url'),
    ('XORG_XCB_SOURCE', 'https://xorg.freedesktop.org/archive/individual/xcb/',
     'xorg xcb source url'),

    # TODO, not urgent, yet nice to have:
    # CPAN_SOURCE GNOME KDE_I18N XCONTRIB DEBIAN KDE GENTOO TEX_CTAN MOZILLA_ALL

    # other constants
    ('SHLIB_EXT', get_shared_lib_ext(), 'extension for shared libraries'),
]

extensions = ['tar.gz', 'tar.xz', 'tar.bz2', 'tgz', 'txz', 'tbz2', 'tb2', 'gtgz', 'zip', 'tar', 'xz', 'tar.Z']
for ext in extensions:
    suffix = ext.replace('.', '_').upper()
    TEMPLATE_CONSTANTS += [
        ('SOURCE_%s' % suffix, '%(name)s-%(version)s.' + ext, "Source .%s bundle" % ext),
        ('SOURCELOWER_%s' % suffix, '%(namelower)s-%(version)s.' + ext, "Source .%s bundle with lowercase name" % ext),
    ]
for pyver in ('py2.py3', 'py2', 'py3'):
    if pyver == 'py2.py3':
        desc = 'Python 2 & Python 3'
        name_infix = ''
    else:
        desc = 'Python ' + pyver[-1]
        name_infix = pyver.upper() + '_'
    TEMPLATE_CONSTANTS += [
        ('SOURCE_%sWHL' % name_infix, '%%(name)s-%%(version)s-%s-none-any.whl' % pyver,
         'Generic (non-compiled) %s wheel package' % desc),
        ('SOURCELOWER_%sWHL' % name_infix, '%%(namelower)s-%%(version)s-%s-none-any.whl' % pyver,
         'Generic (non-compiled) %s wheel package with lowercase name' % desc),
    ]

# TODO derived config templates
# versionmajor, versionminor, versionmajorminor (eg '.'.join(version.split('.')[:2])) )


def template_constant_dict(config, ignore=None, skip_lower=None, toolchain=None):
    """Create a dict for templating the values in the easyconfigs.
        - config is a dict with the structure of EasyConfig._config
    """
    if skip_lower is not None:
        _log.deprecated("Use of 'skip_lower' named argument for template_constant_dict has no effect anymore", '4.0')

    # TODO find better name
    # ignore
    if ignore is None:
        ignore = []
    # make dict
    template_values = {}

    _log.debug("config: %s", config)

    # set 'arch' for system architecture based on 'machine' (4th) element of platform.uname() return value
    template_values['arch'] = platform.uname()[4]

    # step 1: add TEMPLATE_NAMES_EASYCONFIG
    for name in TEMPLATE_NAMES_EASYCONFIG:
        if name in ignore:
            continue

        # check if this template name is already handled
        if template_values.get(name[0]) is not None:
            continue

        if name[0].startswith('toolchain_'):
            tc = config.get('toolchain')
            if tc is not None:
                template_values['toolchain_name'] = tc.get('name', None)
                template_values['toolchain_version'] = tc.get('version', None)
                # only go through this once
                ignore.extend(['toolchain_name', 'toolchain_version'])

        elif name[0].startswith('version_'):
            # parse major and minor version numbers
            version = config['version']
            if version is not None:

                _log.debug("version found in easyconfig is %s", version)
                version = version.split('.')
                try:
                    major = version[0]
                    template_values['version_major'] = major
                    minor = version[1]
                    template_values['version_minor'] = minor
                    template_values['version_major_minor'] = '.'.join([major, minor])
                except IndexError:
                    # if there is no minor version, skip it
                    pass
                # only go through this once
                ignore.extend(['version_major', 'version_minor', 'version_major_minor'])

        elif name[0].endswith('letter'):
            # parse first letters
            if name[0].startswith('name'):
                softname = config['name']
                if softname is not None:
                    template_values['nameletter'] = softname[0]

        elif name[0] == 'module_name':
            template_values['module_name'] = getattr(config, 'short_mod_name', None)

        else:
            raise EasyBuildError("Undefined name %s from TEMPLATE_NAMES_EASYCONFIG", name)

    # step 2: define *ver and *shortver templates
    if TEMPLATE_SOFTWARE_VERSIONS:

        name_to_prefix = dict((name.lower(), pref) for name, pref in TEMPLATE_SOFTWARE_VERSIONS)
        deps = config.get('dependencies', [])

        # also consider build dependencies for *ver and *shortver templates;
        # we need to be a bit careful here, because for iterative installations
        # (when multi_deps is used for example) the builddependencies value may be a list of lists

        # first, determine if we have an EasyConfig instance
        # (indirectly by checking for 'iterating' and 'iterate_options' attributes,
        #  because we can't import the EasyConfig class here without introducing
        #  a cyclic import...);
        # we need to know to determine whether we're iterating over a list of build dependencies
        is_easyconfig = hasattr(config, 'iterating') and hasattr(config, 'iterate_options')
        if is_easyconfig:
            # if we're iterating over different lists of build dependencies,
            # only consider build dependencies when we're actually in iterative mode!
            if 'builddependencies' in config.iterate_options:
                if config.iterating:
                    build_deps = config.get('builddependencies')
                else:
                    build_deps = None
            else:
                build_deps = config.get('builddependencies')
            if build_deps:
                # Don't use += to avoid changing original list
                deps = deps + build_deps
            # include all toolchain deps (e.g. CUDAcore component in fosscuda);
            # access Toolchain instance via _toolchain to avoid triggering initialization of the toolchain!
            if config._toolchain is not None and config._toolchain.tcdeps:
                # If we didn't create a new list above do it here
                if build_deps:
                    deps.extend(config._toolchain.tcdeps)
                else:
                    deps = deps + config._toolchain.tcdeps

        for dep in deps:
            if isinstance(dep, dict):
                dep_name, dep_version = dep['name'], dep['version']

                # take into account dependencies marked as external modules,
                # where name/version may have to be harvested from metadata available for that external module
                if dep.get('external_module', False):
                    metadata = dep.get('external_module_metadata', {})
                    if dep_name is None:
                        # name is a list in metadata, just take first value (if any)
                        dep_name = metadata.get('name', [None])[0]
                    if dep_version is None:
                        # version is a list in metadata, just take first value (if any)
                        dep_version = metadata.get('version', [None])[0]

            elif isinstance(dep, (list, tuple)):
                dep_name, dep_version = dep[0], dep[1]
            else:
                raise EasyBuildError("Unexpected type for dependency: %s", dep)

            if isinstance(dep_name, string_type) and dep_version:
                pref = name_to_prefix.get(dep_name.lower())
                if pref:
                    dep_version = pick_dep_version(dep_version)
                    template_values['%sver' % pref] = dep_version
                    dep_version_parts = dep_version.split('.')
                    template_values['%smajver' % pref] = dep_version_parts[0]
                    if len(dep_version_parts) > 1:
                        template_values['%sminver' % pref] = dep_version_parts[1]
                    template_values['%sshortver' % pref] = '.'.join(dep_version_parts[:2])

    # step 3: add remaining from config
    for name in TEMPLATE_NAMES_CONFIG:
        if name in ignore:
            continue
        if name in config:
            template_values[name] = config[name]
            _log.debug('name: %s, config: %s', name, config[name])

    # step 4. make lower variants
    for name in TEMPLATE_NAMES_LOWER:
        if name in ignore:
            continue

        value = config.get(name) or template_values.get(name)

        if value is None:
            continue
        try:
            template_values[TEMPLATE_NAMES_LOWER_TEMPLATE % {'name': name}] = value.lower()
        except Exception:
            _log.warning("Failed to get .lower() for name %s value %s (type %s)", name, value, type(value))

    # keep track of names of defined templates until now,
    # so we can check whether names of additional dynamic template values are all known
    common_template_names = set(template_values.keys())

    # step 5. add additional conditional templates
    if toolchain is not None and hasattr(toolchain, 'mpi_cmd_prefix'):
        try:
            # get prefix for commands to be run with mpi runtime using default number of ranks
            mpi_cmd_prefix = toolchain.mpi_cmd_prefix()
            if mpi_cmd_prefix is not None:
                template_values['mpi_cmd_prefix'] = mpi_cmd_prefix
        except EasyBuildError as err:
            # don't fail just because we couldn't resolve this template
            _log.warning("Failed to create mpi_cmd_prefix template, error was:\n%s", err)

    # step 6. CUDA compute capabilities
    #         Use the commandline / easybuild config option if given, else use the value from the EC (as a default)
    cuda_compute_capabilities = build_option('cuda_compute_capabilities') or config.get('cuda_compute_capabilities')
    if cuda_compute_capabilities:
        template_values['cuda_compute_capabilities'] = ','.join(cuda_compute_capabilities)
        template_values['cuda_cc_space_sep'] = ' '.join(cuda_compute_capabilities)
        template_values['cuda_cc_semicolon_sep'] = ';'.join(cuda_compute_capabilities)
        sm_values = ['sm_' + cc.replace('.', '') for cc in cuda_compute_capabilities]
        template_values['cuda_sm_comma_sep'] = ','.join(sm_values)
        template_values['cuda_sm_space_sep'] = ' '.join(sm_values)

    unknown_names = []
    for key in template_values:
        dynamic_template_names = set(x for (x, _) in TEMPLATE_NAMES_DYNAMIC)
        if not (key in common_template_names or key in dynamic_template_names):
            unknown_names.append(key)
    if unknown_names:
        raise EasyBuildError("One or more template values found with unknown name: %s", ','.join(unknown_names))

    return template_values


def to_template_str(key, value, templ_const, templ_val):
    """
    Insert template values where possible

    :param key: name of easyconfig parameter
    :param value: string representing easyconfig parameter value
    :param templ_const: dictionary of template strings (constants)
    :param templ_val: (ordered) dictionary of template strings specific for this easyconfig file
    """
    old_value = None
    while value != old_value:
        old_value = value
        # check for constant values
        for tval, tname in templ_const.items():
            if tval in value:
                value = re.sub(r'(^|\W)' + re.escape(tval) + r'(\W|$)', r'\1' + tname + r'\2', value)

        for tval, tname in templ_val.items():
            # only replace full words with templates: word to replace should be at the beginning of a line
            # or be preceded by a non-alphanumeric (\W). It should end at the end of a line or be succeeded
            # by another non-alphanumeric;
            # avoid introducing self-referencing easyconfig parameter value
            # by taking into account given name of easyconfig parameter ('key')
            if tval in value and tname != key:
                value = re.sub(r'(^|\W)' + re.escape(tval) + r'(\W|$)', r'\1%(' + tname + r')s\2', value)

            # special case of %(pyshortver)s, where we should template 'python2.7' to 'python%(pyshortver)s'
            if tname == 'pyshortver' and ('python' + tval) in value:
                value = re.sub(r'(^|\W)python' + re.escape(tval) + r'(\W|$)', r'\1python%(' + tname + r')s\2', value)

    return value


def template_documentation():
    """Generate the templating documentation"""
    # This has to reflect the methods/steps used in easyconfig _generate_template_values
    indent_l0 = " " * 2
    indent_l1 = indent_l0 + " " * 2
    doc = []

    # step 1: add TEMPLATE_NAMES_EASYCONFIG
    doc.append('Template names/values derived from easyconfig instance')
    for name in TEMPLATE_NAMES_EASYCONFIG:
        doc.append("%s%%(%s)s: %s" % (indent_l1, name[0], name[1]))

    # step 2: add *ver/*shortver templates for software listed in TEMPLATE_SOFTWARE_VERSIONS
    doc.append("Template names/values for (short) software versions")
    for name, pref in TEMPLATE_SOFTWARE_VERSIONS:
        doc.append("%s%%(%sshortver)s: short version for %s (<major>.<minor>)" % (indent_l1, pref, name))
        doc.append("%s%%(%sver)s: full version for %s" % (indent_l1, pref, name))

    # step 3: add remaining self._config
    doc.append('Template names/values as set in easyconfig')
    for name in TEMPLATE_NAMES_CONFIG:
        doc.append("%s%%(%s)s" % (indent_l1, name))

    # step 4. make lower variants
    doc.append('Lowercase values of template values')
    for name in TEMPLATE_NAMES_LOWER:
        namelower = TEMPLATE_NAMES_LOWER_TEMPLATE % {'name': name}
        doc.append("%s%%(%s)s: lower case of value of %s" % (indent_l1, namelower, name))

    # step 5. self.template_values can/should be updated from outside easyconfig
    # (eg the run_setp code in EasyBlock)
    doc.append('Template values set outside EasyBlock runstep')
    for name in TEMPLATE_NAMES_EASYBLOCK_RUN_STEP:
        doc.append("%s%%(%s)s: %s" % (indent_l1, name[0], name[1]))

    doc.append('Template constants that can be used in easyconfigs')
    for cst in TEMPLATE_CONSTANTS:
        doc.append('%s%s: %s (%s)' % (indent_l1, cst[0], cst[2], cst[1]))

    return "\n".join(doc)



# #
# Copyright 2012-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
Module that contains a set of classes and function to generate variables to be used
e.g., in compiling or linking

:author: Stijn De Weirdt (Ghent University)
:author: Kenneth Hoste (Ghent University)
"""
import copy
import os

from easybuild.base import fancylogger
from easybuild.tools.build_log import EasyBuildError


_log = fancylogger.getLogger('variables', fname=False)


def get_class(name, default_class, map_class=None):
    """Return class based on default
        map_class
             if key == str -> value = class
             else: key = class -> list of strings
    """
    if map_class is None:
        map_class = {}

    klass = default_class
    if name is not None:
        try:
            klass = map_class[name]
        except BaseException:
            for k, v in map_class.items():
                if type(k) in (type,) and name in v:
                    klass = k
                    break

    return klass


def join_map_class(map_classes):
    """Join all class_maps into single class_map"""
    res = {}
    for map_class in map_classes:
        for key, val in map_class.items():
            if isinstance(key, (str,)):
                var_name = key
                if isinstance(val, (tuple, list)):
                    # second element is documentation
                    klass = val[0]
                res[var_name] = klass
            elif type(key) in (type,):
                # k is the class, v a list of tuples (name,doc)
                klass = key
                default = res.setdefault(klass, [])
                default.extend([tpl[0] for tpl in val])
            else:
                raise EasyBuildError("join_map_class: impossible to join key %s value %s", key, val)

    return res


class StrList(list):
    """List of strings"""
    SEPARATOR = ' '

    PREFIX = None
    SUFFIX = None

    BEGIN = None
    END = None

    POSITION = 0  # when sorting in list of list: < 0 -> left; > 0 : right
    SANITIZE_REMOVE_DUPLICATE_KEEP = None  # used with ListOfList

    JOIN_BEGIN_END = False

    def __init__(self, *args, **kwargs):
        super(StrList, self).__init__(*args, **kwargs)
        self.log = fancylogger.getLogger(self.__class__.__name__, fname=False)

    def str_convert(self, x):
        """Convert members of list to string (no prefix of begin and end)"""
        return ''.join([str(y) for y in [self.PREFIX, str(x), self.SUFFIX] if y is not None])

    def _str_ok(self, x):
        """Test if x can be added to returned string"""
        test = x is not None and len(str(x)) > 0
        return test

    def _str_self(self):
        """Main part of __str__"""
        return [self.str_convert(x) for x in self if self._str_ok(x)]

    def sanitize(self):
        """Sanitize self"""

    def __str__(self):
        """_str_self and support for BEGIN/END"""
        self.sanitize()
        xs = [self.BEGIN] + self._str_self() + [self.END]
        return str(self.SEPARATOR).join([str(x) for x in xs if self._str_ok(x)])

    def __getattribute__(self, attr_name):
        """Filter out function calls from Variables class"""
        if attr_name == 'nappend_el':
            return self.append
        elif attr_name == 'nextend_el':
            return self.extend
        else:
            return super(StrList, self).__getattribute__(attr_name)

    def copy(self):
        """Return copy of self"""
        return copy.deepcopy(self)

    def try_remove(self, values):
        """Remove without ValueError in case of missing element"""
        for value in values:
            try:
                self.remove(value)
            except ValueError:
                pass


class CommaList(StrList):
    """Comma-separated list"""
    SEPARATOR = ','


class AbsPathList(StrList):
    """Absolute paths (eg -L or -I)"""

    SANITIZE_REMOVE_DUPLICATE_KEEP = -1  # sanitize from end

    def append_exists(self, prefix, paths, suffix=None, filename=None, append_all=False):
        """
        Given prefix and list of paths, return first that exists
            if suffix : extend the paths with prefixes
            if filename : look for filename in prefix+paths
        """
        self.log.devel("append_exists: prefix %s paths %s suffix %s filename %s append_all %s",
                       prefix, paths, suffix, filename, append_all)
        if suffix is not None:
            res = []
            for path in paths:
                res.extend(["%s%s" % (path, suffix), path])
            paths = res

        for path in paths:
            abs_path = os.path.join(prefix, path)
            if filename is not None:
                abs_path = os.path.join(abs_path, filename)
            if os.path.exists(abs_path):
                self.append(abs_path)
                self.log.devel("append_exists: added absolute path %s", abs_path)
                if not append_all:
                    return

    def append_subdirs(self, base, subdirs=None):
        """
        Add directory base, or its subdirs if subdirs is not None
        """
        self.log.devel("append_subdirs: base %s subdirs %s", base, subdirs)

        if subdirs is None:
            subdirs = [None]
        for subdir in subdirs:
            if subdir is None:
                directory = base
            else:
                directory = os.path.join(base, subdir)

            if os.path.isdir(directory):
                self.append(directory)
                self.log.devel("append_subdirs: added directory %s", directory)
            else:
                self.log.warning("flags_for_subdirs: directory %s was not found" % directory)


class ListOfLists(list):
    """List of lists"""
    DEFAULT_CLASS = StrList
    PROTECTED_CLASSES = []  # classes that are not converted to DEFAULT_CLASS
    # PROTECTED_INSTANCES = [AbsPathList, LibraryList]
    PROTECTED_INSTANCES = []
    PROTECT_CLASS_SELF = True  # don't convert values that are same class as DEFAULT_CLASS
    PROTECT_INSTANCE_SELF = True  # don't convert values that are instance of DEFAULT_CLASS

    SEPARATOR = None

    SANITIZE_SORT = True
    SANITIZE_REMOVE_DUPLICATE = False
    SANITIZE_REMOVE_DUPLICATE_KEEP = None

    JOIN_BEGIN_END = False

    def __init__(self, *args, **kwargs):
        super(ListOfLists, self).__init__(*args, **kwargs)
        self.log = fancylogger.getLogger(self.__class__.__name__, fname=False)
        self._first = None

        self.protected_classes = self.PROTECTED_CLASSES[:]
        if self.PROTECT_CLASS_SELF:
            if self.DEFAULT_CLASS not in self.protected_classes:
                self.protected_classes.append(self.DEFAULT_CLASS)
        self.protected_instances = self.PROTECTED_INSTANCES[:]
        if self.PROTECT_INSTANCE_SELF:
            if self.DEFAULT_CLASS not in self.protected_instances:
                self.protected_instances.append(self.DEFAULT_CLASS)

    def append_empty(self):
        """Initialise MAP_CLASS instance"""
        self.nappend(None, append_empty=True)

    def show_el(self):
        """Show some info on the elements"""
        res = []
        for el in self:
            res.append("%s B_ %s _E" % (type(el), el))
        return ";".join(res)

    def get_first(self):
        """Return first non-empty list
            if it doesn't exist, try to return first element
        """
        for x in self:
            if self._str_ok(x):
                return x

        if len(self) > 0:
            return self[0]

    def _is_protected(self, value):
        """Check if value is protected from conversion to default class"""
        res = False

        if type(value) in self.protected_classes:
            self.log.devel("_is_protected: %s value %s (%s)", self.protected_classes, value, type(value))
            res = True
        elif isinstance(value, tuple(self.protected_instances)):
            self.log.devel("_is_protected: %s value %s (%s)", self.protected_instances, value, type(value))
            res = True

        self.log.devel("_is_protected: %s value %s (%s)", res, value, value.__repr__())
        return res

    def nappend(self, value, **kwargs):
        """Named append
            name is not used anymore
        """
        append_empty = kwargs.pop('append_empty', False)
        position = kwargs.pop('position', None)
        klass = kwargs.pop('var_class', self.DEFAULT_CLASS)

        if self._is_protected(value):
            newvalue = value.copy()
        else:
            if isinstance(value, (str, int,)):
                # convert to list. although the try/except will work
                #  list('XYZ') creates ['X','Y','Z']
                value = [value]

            try:
                # this might work, but probably not
                newvalue = klass(value, **kwargs)
            except BaseException:
                newvalue = klass(**kwargs)
                if value is not None:
                    newvalue.append(value)
        if position is not None:
            newvalue.POSITION = position
        if self._str_ok(newvalue) or append_empty:
            self.append(newvalue)
            self.log.devel("nappend: value %s newvalue %s position %s",
                           value.__repr__(), newvalue.__repr__(), position)
            return newvalue
        else:
            self.log.devel("nappend: ignoring value %s newvalue %s (not _str_ok)",
                           value.__repr__(), newvalue.__repr__())

    def nextend(self, value=None, **kwargs):
        """Named extend, value is list type (TODO: tighten the allowed values)
            name not used anymore
        """
        klass = kwargs.pop('var_class', self.DEFAULT_CLASS)
        res = []
        if value is None:
            # TODO ? append_empty ?
            raise EasyBuildError("extend_el with None value unimplemented")
        else:
            for el in value:
                if not self._str_ok(el):
                    self.log.devel("nextend: ignoring el %s from value %s (not _str_ok)", el, value.__repr__())
                    continue

                if type(el) in self.PROTECTED_CLASSES:
                    newvalue = el
                else:
                    if isinstance(el, (str, int,)):
                        # convert to list. although the try/except will work
                        #  list('XYZ') creates ['X','Y','Z']
                        el = [el]

                    try:
                        # this might work, but probably not
                        newvalue = klass(el)
                    except BaseException:
                        newvalue = klass()
                        if value is not None:
                            newvalue.append(el)

                res.append(newvalue)

        self.extend(res)
        self.log.devel("nextend: value %s res %s", value.__repr__(), res.__repr__())
        return res

    def str_convert(self, x):
        """Given x, return a string representing x
            called in __str__ of this class
        """
        return str(x)

    def _str_ok(self, x):
        """Test if x can be added returned string"""
        test = x is not None and len(x) > 0
        return test

    def sanitize(self):
        """Cleanup self"""
        if self.SANITIZE_SORT:
            self.sort(key=lambda x: getattr(x, 'POSITION'))

        if self.SANITIZE_REMOVE_DUPLICATE:
            # get all occurences with their index
            to_remove = []
            for el in self:
                all_idx = [idx for idx, x in enumerate(self) if x == el]
                if len(all_idx) > 1:
                    if self.SANITIZE_REMOVE_DUPLICATE_KEEP == 0:
                        # keep first
                        to_remove.extend(all_idx[1:])
                    elif self.SANITIZE_REMOVE_DUPLICATE_KEEP == -1:
                        # keep last
                        to_remove.extend(all_idx[:-1])

            to_remove = sorted(list(set(to_remove)), reverse=True)
            self.log.devel("sanitize: to_remove in %s %s", self.__repr__(), to_remove)
            for idx in to_remove:
                del self[idx]

        if self.JOIN_BEGIN_END:
            # group elements with same begin/end into one element
            to_remove = []
            # work in reversed order; don't check last one (ie real el 0), it has no next element
            for idx in range(1, len(self))[::-1]:
                if self[idx].BEGIN is None or self[idx].END is None:
                    continue
                self.log.devel("idx %s len %s", idx, len(self))
                # do check POSITION, sorting already done
                if self[idx].BEGIN == self[idx - 1].BEGIN and self[idx].END == self[idx - 1].END:
                    self.log.devel("sanitize: JOIN_BEGIN_END idx %s joining %s and %s", idx, self[idx], self[idx - 1])
                    self[idx - 1].extend(self[idx])
                    to_remove.append(idx)  # remove current el
            to_remove = sorted(list(set(to_remove)), reverse=True)
            for idx in to_remove:
                del self[idx]

    def flatten(self):
        res = []
        for x in self:
            if self._str_ok(x):
                res.extend(x)
        return res

    def __str__(self):
        self._first = self.get_first()
        self.sanitize()
        sep = ''  # default no separator

        if self._first is None:
            # return empty string
            self.log.devel("__str__: first is None (self %s)", self.__repr__())
            return ''
        else:
            sep = self.SEPARATOR

            listoflists = self
            if isinstance(self._first, AbsPathList):
                # reverse list to mimic prepend_paths in module files
                listoflists = reversed(self)
            txt = str(sep).join([self.str_convert(x) for x in listoflists if self._str_ok(x)])
            self.log.devel("__str__: return %s (self: %s)", txt, self.__repr__())
            return txt

    def try_function_on_element(self, function_name, names=None, args=None, kwargs=None):
        """Try to run function function_name on each element"""
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        for el in self:
            if hasattr(el, function_name):
                function = getattr(el, function_name)
                function(*args, **kwargs)

    def try_remove(self, values):
        """Try to remove one or more values from the elements"""
        self.try_function_on_element('try_remove', args=[values])

    def copy(self):
        """Return copy of self"""
        return copy.deepcopy(self)


class Variables(dict):
    """
    Class to hold variable-like key/value pairs
        All values are lists (or derived from list class)
            most only have a single element though
            some are lists of lists
        __str__ creates a single string

        Most items are of same DEFAULT_CLASS
            but are in different classes
    """
    DEFAULT_LISTCLASS = ListOfLists
    MAP_LISTCLASS = {}  # map between variable name and ListOfList classes (ie not the (default) class for the variable)

    DEFAULT_CLASS = StrList
    MAP_CLASS = {}  # predefined map to specify (default) mapping between variables and classes

    def __init__(self, *args, **kwargs):
        super(Variables, self).__init__(*args, **kwargs)
        self.log = fancylogger.getLogger(self.__class__.__name__, fname=False)

    def get_list_class(self, name):
        """Return the class associated with the name according to the DEFAULT_LISTCLASS and MAP_LISTCLASS"""
        return get_class(name, self.DEFAULT_LISTCLASS, self.MAP_LISTCLASS)

    def get_element_class(self, name):
        """Return the class associated with the name according to the DEFAULT_CLASS and MAP_CLASS"""
        return get_class(name, self.DEFAULT_CLASS, self.MAP_CLASS)

    def get_instance(self, name=None):
        """Return an instance of the class"""
        list_class = self.get_list_class(name)
        element_class = self.get_element_class(name)

        class klass(list_class):
            DEFAULT_CLASS = element_class

            SEPARATOR = element_class.SEPARATOR

            SANITIZE_REMOVE_DUPLICATE = element_class.SANITIZE_REMOVE_DUPLICATE_KEEP is not None
            SANITIZE_REMOVE_DUPLICATE_KEEP = element_class.SANITIZE_REMOVE_DUPLICATE_KEEP

            JOIN_BEGIN_END = element_class.JOIN_BEGIN_END

        # better log messages (most use self.__class__.__name__; would give klass otherwise)
        klass.__name__ = "%s_%s" % (self.__class__.__name__, name)
        return klass()

    def join(self, name, *others):
        """Join all values in others into name
            it is first tested if other is an existing element
                else it is nappend-ed
        """
        self.log.devel("join name %s others %s", name, others)

        # make sure name is defined, even if 'others' list is empty
        self.setdefault(name)

        for other in others:
            if other in self:
                self.log.devel("join other %s in self: other %s", other, self.get(other).__repr__())
                for el in self.get(other):
                    self.nappend(name, el)
            else:
                raise EasyBuildError("join: name %s; other %s not found in self.", name, other)

    def append(self, name, value):
        """Append value to element name (alias for nappend)"""
        return self.nappend(name, value)

    def __setitem__(self, name, value):
        """Automatically creates a list for each name"""
        if name in self:
            del self[name]
        self.nappend(name, value)

    def setdefault(self, name, default=None, append_empty=False):
        # """append_empty to non-existing element"""
        if name in self:
            default = self[name]
        else:
            if default is None:
                default = self.get_instance(name)
            super(Variables, self).__setitem__(name, default)

        if len(default) == 0:
            self.log.devel("setdefault: name %s initialising.", name)
            if append_empty:
                default.append_empty()
        return default

    def try_function_on_element(self, function_name, names=None, args=None, kwargs=None):
        """Try to run function function_name on each element of names"""
        if names is None:
            names = self.keys()
        for name in names:
            self.log.devel("try_function_el: name %s function_name %s", name, function_name)
            self[name].try_function_on_element(function_name, args=args, kwargs=kwargs)

    def __getattribute__(self, attr_name):
        # allow for pass-through
        if attr_name in ['nappend', 'nextend', 'append_empty', 'first', 'get_class']:
            self.log.devel("Passthrough to LISTCLASS function %s", attr_name)

            def _passthrough(name, *args, **kwargs):
                """functions that pass through to LISTCLASS instances"""
                current = self.setdefault(name)
                actual_function = getattr(current, attr_name)
                res = actual_function(*args, **kwargs)
                return res
            return _passthrough
        elif attr_name in ['nappend_el', 'nextend_el', 'append_exists', 'append_subdirs']:
            self.log.devel("Passthrough to LISTCLASS element function %s", attr_name)

            def _passthrough(name, *args, **kwargs):
                """"Functions that pass through to elements of LISTCLASS (accept idx as index)"""
                idx = kwargs.pop('idx', -1)
                if attr_name in ['append_exists', 'append_subdirs']:
                    current = self.setdefault(name)
                    current.append_empty()  # always add empty
                else:
                    current = self.setdefault(name, append_empty=True)
                actual_function = getattr(current[idx], attr_name)
                res = actual_function(*args, **kwargs)
                return res
            return _passthrough
        else:
            return super(Variables, self).__getattribute__(attr_name)



##
# Copyright 2014-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
Support for Cray's LibSci library, which provides BLAS/LAPACK support.
cfr. https://www.nersc.gov/users/software/programming-libraries/math-libraries/libsci/

:author: Petar Forai (IMP/IMBA, Austria)
:author: Kenneth Hoste (Ghent University)
"""
import os

from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.toolchain.linalg import LinAlg


CRAY_LIBSCI_MODULE_NAME = 'cray-libsci'
TC_CONSTANT_CRAY_LIBSCI = 'CrayLibSci'


class LibSci(LinAlg):
    """Support for Cray's LibSci library, which provides BLAS/LAPACK support."""
    # BLAS/LAPACK support
    # via cray-libsci module, which gets loaded via the PrgEnv module
    # see https://www.nersc.gov/users/software/programming-libraries/math-libraries/libsci/
    BLAS_MODULE_NAME = [CRAY_LIBSCI_MODULE_NAME]

    # no need to specify libraries, compiler driver takes care of linking the right libraries
    # FIXME: need to revisit this, on numpy we ended up with a serial BLAS through the wrapper.
    BLAS_LIB = ['']
    BLAS_LIB_MT = ['']
    BLAS_FAMILY = TC_CONSTANT_CRAY_LIBSCI

    LAPACK_MODULE_NAME = [CRAY_LIBSCI_MODULE_NAME]
    LAPACK_IS_BLAS = True
    LAPACK_FAMILY = TC_CONSTANT_CRAY_LIBSCI

    BLACS_MODULE_NAME = []
    SCALAPACK_MODULE_NAME = []

    def _get_software_root(self, name, required=True):
        """Get install prefix for specified software name; special treatment for Cray modules."""
        if name == 'cray-libsci':
            # Cray-provided LibSci module
            env_var = 'CRAY_LIBSCI_PREFIX_DIR'
            root = os.getenv(env_var, None)
            if root is None:
                if required:
                    raise EasyBuildError("Failed to determine install prefix for %s via $%s", name, env_var)
            else:
                self.log.debug("Obtained install prefix for %s via $%s: %s", name, env_var, root)
        else:
            root = super(LibSci, self)._get_software_root(name, required=required)

        return root

    def _set_blacs_variables(self):
        """Skip setting BLACS related variables"""
        pass

    def _set_scalapack_variables(self):
        """Skip setting ScaLAPACK related variables"""
        pass

    def definition(self):
        """
        Filter BLAS module from toolchain definition.
        The cray-libsci module is loaded indirectly (and versionless) via the PrgEnv module,
        and thus is not a direct toolchain component.
        """
        tc_def = super(LibSci, self).definition()
        tc_def['BLAS'] = []
        tc_def['LAPACK'] = []
        return tc_def



#
# Copyright 2013-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Easyconfig licenses module that provides all licenses that can
be used within an Easyconfig file.

:author: Stijn De Weirdt (Ghent University)
:author: Kenneth Hoste (Ghent University)
"""

from easybuild.base import fancylogger
from easybuild.tools.utilities import get_subclasses


_log = fancylogger.getLogger('easyconfig.licenses', fname=False)


class License(object):
    """EasyBuild easyconfig license class
        This is also the default restrictive license
    """
    HIDDEN = False  # disable subclasses from being seen/used
    NAME = None
    VERSION = None
    DESCRIPTION = None

    DISTRIBUTE_SOURCE = False  # does the license allows to (re)distribute the code
    GROUP_SOURCE = True  # does the license require to keep the source under dedicated group
    GROUP_BINARY = True  # does the license require to install the binaries under dedicated group

    CLASSNAME_PREFIX = 'License'

    @property
    def name(self):
        """Return license name."""
        if self.NAME is None:
            name = self.__class__.__name__
            if name.startswith(self.CLASSNAME_PREFIX):
                name = name[len(self.CLASSNAME_PREFIX):]
        else:
            name = self.NAME

        return name

    def __init__(self):
        """License constructor."""
        self.version = self.VERSION
        self.description = self.DESCRIPTION
        self.distribute_source = self.DISTRIBUTE_SOURCE
        self.group_source = self.GROUP_SOURCE
        self.group_binary = self.GROUP_BINARY


class LicenseVeryRestrictive(License):
    """Default license should be very restrictive, so nothing to do here, just a placeholder"""
    pass


class LicenseUnknown(LicenseVeryRestrictive):
    """A (temporary) license, could be used as default in case nothing was specified"""
    pass


# inspiration
# http://en.wikipedia.org/wiki/Category:Free_and_open-source_software_licenses

class LicenseOpen(License):
    """
    Hidden license class to subclass open licenses.
    'Open' here means, that source can be redistributed, and that both source
    and binaries do not need special groups (ie anyone can access/use it).
    """
    HIDDEN = True
    DISTRIBUTE_SOURCE = True
    GROUP_SOURCE = False
    GROUP_BINARY = False


class LicenseGPL(LicenseOpen):
    """
    Hidden license class to subclass GPL licenses.
    """
    DESCRIPTION = ("The GNU General Public License is a free, "
                   "copyleft license for software and other kinds of works.")


class LicenseGPLv2(LicenseGPL):
    """GPLv2 license"""
    HIDDEN = False
    VERSION = (2,)


class LicenseGPLv3(LicenseGPLv2):
    """GPLv3 license"""
    VERSION = (3,)


class LicenseGCC(LicenseGPLv3):
    """GPLv3 with GCC Runtime Library Exception.
        Latest GPLv2 GCC release was 4.2.1 (http://gcc.gnu.org/ml/gcc-announce/2007/msg00003.html).
    """
    DESCRIPTION = ("The GNU General Public License is a free, "
                   "copyleft license for software and other kinds of works. "
                   "The GCC Runtime Library Exception is an additional permission "
                   "under section 7 of the GNU General Public License, version 3.")


class LicenseGCCOld(LicenseGPLv2):
    """GPLv2 with GCC Runtime Library Exception for older GCC versions.
        Latest GPLv2 GCC release was 4.2.1 (http://gcc.gnu.org/ml/gcc-announce/2007/msg00003.html).
    """
    DESCRIPTION = LicenseGCC.DESCRIPTION


class LicenseZlib(LicenseOpen):
    """The zlib License is a permissive free software license
        http://www.zlib.net/zlib_license.html
    """
    DESCRIPTION = ("Permission is granted to anyone to use this software for any purpose,"
                   " including commercial applications, and to alter it and redistribute it"
                   " freely, subject to 3 restrictions;"
                   " http://www.zlib.net/zlib_license.html for full license")


class LicenseLibpng(LicenseOpen):
    """The PNG license is derived from the zlib license,
        http://libpng.org/pub/png/src/libpng-LICENSE.txt
    """
    HIDDEN = False
    DESCRIPTION = ("Permission is granted to use, copy, modify, and distribute the "
                   "source code, or portions hereof, for any purpose, without fee, subject "
                   "to 3 restrictions; http://libpng.org/pub/png/src/libpng-LICENSE.txt for full license")


def what_licenses():
    """Return a dict of License subclasses names and license instances"""
    res = {}
    for lic in get_subclasses(License):
        if lic.HIDDEN:
            continue
        res[lic.__name__] = lic

    return res


EASYCONFIG_LICENSES_DICT = what_licenses()



# #
# Copyright 2013-2022 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #

"""
The main easyconfig format class

:author: Stijn De Weirdt (Ghent University)
:author: Kenneth Hoste (Ghent University)
"""
import copy
import re
import sys

from easybuild.base import fancylogger
from easybuild.framework.easyconfig.constants import EASYCONFIG_CONSTANTS
from easybuild.framework.easyconfig.format.format import get_format_version, EasyConfigFormat
from easybuild.framework.easyconfig.licenses import EASYCONFIG_LICENSES_DICT
from easybuild.framework.easyconfig.templates import TEMPLATE_CONSTANTS
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.configobj import ConfigObj
from easybuild.tools.systemtools import get_shared_lib_ext


_log = fancylogger.getLogger('easyconfig.format.pyheaderconfigobj', fname=False)


def build_easyconfig_constants_dict():
    """Make a dictionary with all constants that can be used"""
    all_consts = [
        ('TEMPLATE_CONSTANTS', dict([(x[0], x[1]) for x in TEMPLATE_CONSTANTS])),
        ('EASYCONFIG_CONSTANTS', dict([(key, val[0]) for key, val in EASYCONFIG_CONSTANTS.items()])),
        ('EASYCONFIG_LICENSES', dict([(klass().name, name) for name, klass in EASYCONFIG_LICENSES_DICT.items()])),
    ]
    err = []
    const_dict = {}

    for (name, csts) in all_consts:
        for cst_key, cst_val in csts.items():
            ok = True
            for (other_name, other_csts) in all_consts:
                if name == other_name:
                    continue
                # make sure that all constants only belong to one name
                if cst_key in other_csts:
                    err.append('Found name %s from %s also in %s' % (cst_key, name, other_name))
                    ok = False
            if ok:
                const_dict[cst_key] = cst_val

    if len(err) > 0:
        raise EasyBuildError("EasyConfig constants sanity check failed: %s", '\n'.join(err))
    else:
        return const_dict


def build_easyconfig_variables_dict():
    """Make a dictionary with all variables that can be used"""
    vars_dict = {
        "shared_lib_ext": get_shared_lib_ext(),
    }

    return vars_dict


class EasyConfigFormatConfigObj(EasyConfigFormat):
    """
    Extended EasyConfig format, with support for a header and sections that are actually parsed (as opposed to exec'ed).
    It's very limited for now, but is already huge improvement.

    3 parts in easyconfig file:
    - header (^# style)
    - pyheader (including docstring)
     - contents is exec'ed, docstring and remainder are extracted
    - begin of regular section until EOF
     - feed to ConfigObj
    """

    PYHEADER_ALLOWED_BUILTINS = []  # default no builtins
    PYHEADER_MANDATORY = None  # no defaults
    PYHEADER_BLACKLIST = None  # no defaults

    def __init__(self, *args, **kwargs):
        """Extend EasyConfigFormat with some more attributes"""
        super(EasyConfigFormatConfigObj, self).__init__(*args, **kwargs)
        self.pyheader_localvars = None
        self.configobj = None

    def parse(self, txt, strict_section_markers=False):
        """
        Pre-process txt to extract header, docstring and pyheader
        Then create the configobj instance by parsing the remainder
        """
        # where is the first section?
        sectionmarker_pattern = ConfigObj._sectionmarker.pattern
        if strict_section_markers:
            # don't allow indentation for section markers
            # done by rewriting section marker regex, such that we don't have to patch configobj.py
            indented_markers_regex = re.compile('^.*?indentation.*$', re.M)
            sectionmarker_pattern = indented_markers_regex.sub('', sectionmarker_pattern)
        regex = re.compile(sectionmarker_pattern, re.VERBOSE | re.M)
        reg = regex.search(txt)
        if reg is None:
            # no section
            self.log.debug("No section found.")
            start_section = None
        else:
            start_section = reg.start()
            last_n = 100
            pre_section_tail = txt[start_section - last_n:start_section]
            sections_head = txt[start_section:start_section + last_n]
            self.log.debug('Sections start at index %s, %d-chars context:\n"""%s""""\n<split>\n"""%s..."""',
                           start_section, last_n, pre_section_tail, sections_head)

        self.parse_pre_section(txt[:start_section])
        if start_section is not None:
            self.parse_section_block(txt[start_section:])

    def parse_pre_section(self, txt):
        """Parse the text block before the start of the first section"""
        header_text = []
        txt_list = txt.split('\n')
        header_reg = re.compile(r'^\s*(#.*)?$')

        # pop lines from txt_list into header_text, until we're not in header anymore
        while len(txt_list) > 0:
            line = txt_list.pop(0)

            format_version = get_format_version(line)
            if format_version is not None:
                if not format_version == self.VERSION:
                    raise EasyBuildError("Invalid format version %s for current format class", format_version)
                else:
                    self.log.info("Valid format version %s found" % format_version)
                # version is not part of header
                continue

            r = header_reg.search(line)
            if not r:
                # put the line back, and quit (done with header)
                txt_list.insert(0, line)
                break
            header_text.append(line)

        self.parse_header('\n'.join(header_text))
        self.parse_pyheader('\n'.join(txt_list))

    def parse_header(self, header):
        """Parse the header, assign to self.header"""
        # FIXME: do something with the header
        self.log.debug("Found header %s" % header)
        self.header = header

    def parse_pyheader(self, pyheader):
        """Parse the python header, assign to docstring and cfg"""
        global_vars = self.pyheader_env()
        self.log.debug("pyheader initial global_vars %s", global_vars)
        self.log.debug("pyheader text being exec'ed: %s", pyheader)

        # check for use of deprecated magic easyconfigs variables
        for magic_var in build_easyconfig_variables_dict():
            if re.search(magic_var, pyheader, re.M):
                _log.nosupport("Magic 'global' easyconfigs variable %s should no longer be used" % magic_var, '2.0')

        # copy dictionary with constants that can be used in easyconfig files,
        # use it as 'globals' dict in exec call so parsed easyconfig parameters are added to it
        cfg = copy.deepcopy(global_vars)

        try:
            # cfg dict is used as globals dict;
            # we should *not* pass a separate (empty) locals dict to exec,
            # otherwise problems may occur when using Python 3 and
            # parsing easyconfig files that use local variables in list comprehensions
            # cfr. https://github.com/easybuilders/easybuild-framework/pull/2895
            exec(pyheader, cfg)
        except Exception as err:  # pylint: disable=broad-except
            err_msg = str(err)
            exc_tb = sys.exc_info()[2]
            if exc_tb.tb_next is not None:
                err_msg += " (line %d)" % exc_tb.tb_next.tb_lineno
            raise EasyBuildError("Parsing easyconfig file failed: %s", err_msg)

        self.log.debug("pyheader parsed cfg: %s", cfg)

        # get rid of constants from parsed easyconfig file, they are not valid easyconfig parameters
        for key in global_vars:
            self.log.debug("Removing key '%s' from parsed cfg (constant, not an easyconfig parameter)", key)
            del cfg[key]

        self.log.debug("pyheader final parsed cfg: %s", cfg)

        if '__doc__' in cfg:
            self.docstring = cfg.pop('__doc__')
        else:
            self.log.debug('No docstring found in cfg')

        self.pyheader_localvars = cfg

    def pyheader_env(self):
        """Create the global/local environment to use with eval/execfile"""
        global_vars = {}

        # all variables
        global_vars.update(build_easyconfig_variables_dict())
        # all constants
        global_vars.update(build_easyconfig_constants_dict())

        # allowed builtins
        if self.PYHEADER_ALLOWED_BUILTINS is not None:
            current_builtins = globals()['__builtins__']
            builtins = {}
            for name in self.PYHEADER_ALLOWED_BUILTINS:
                if hasattr(current_builtins, name):
                    builtins[name] = getattr(current_builtins, name)
                elif isinstance(current_builtins, dict) and name in current_builtins:
                    builtins[name] = current_builtins[name]
                else:
                    self.log.warning('No builtin %s found.' % name)
            global_vars['__builtins__'] = builtins
            self.log.debug("Available builtins: %s" % global_vars['__builtins__'])

        return global_vars

    def _validate_pyheader(self):
        """
        Basic validation of pyheader localvars.
        This takes parameter names from the PYHEADER_BLACKLIST and PYHEADER_MANDATORY;
        blacklisted parameters are not allowed, mandatory parameters are mandatory unless blacklisted
        """
        if self.pyheader_localvars is None:
            raise EasyBuildError("self.pyheader_localvars must be initialized")
        if self.PYHEADER_BLACKLIST is None or self.PYHEADER_MANDATORY is None:
            raise EasyBuildError('Both PYHEADER_BLACKLIST and PYHEADER_MANDATORY must be set')

        for param in self.PYHEADER_BLACKLIST:
            if param in self.pyheader_localvars:
                # TODO add to easyconfig unittest (similar to mandatory)
                raise EasyBuildError('blacklisted param %s not allowed in pyheader', param)

        missing = []
        for param in self.PYHEADER_MANDATORY:
            if param in self.PYHEADER_BLACKLIST:
                continue
            if param not in self.pyheader_localvars:
                missing.append(param)
        if missing:
            raise EasyBuildError('mandatory parameters not provided in pyheader: %s', ', '.join(missing))

    def parse_section_block(self, section):
        """Parse the section block by trying to convert it into a ConfigObj instance"""
        try:
            self.configobj = ConfigObj(section.split('\n'))
        except SyntaxError as err:
            raise EasyBuildError('Failed to convert section text %s: %s', section, err)

        self.log.debug("Found ConfigObj instance %s" % self.configobj)
